@@ -7,16 +7,20 @@
 processes in a portable way by using Python.
 """
 
-__version__ = "0.2.0"
+__version__ = "0.2.1"
 version_info = tuple([int(num) for num in __version__.split('.')])
 
 __all__ = [
     # exceptions
-    "Error", "NoSuchProcess", "AccessDenied",
+    "Error", "NoSuchProcess", "AccessDenied", "TimeoutExpired",
     # constants
-    "NUM_CPUS", "TOTAL_PHYMEM", "version_info", "__version__",
+    "NUM_CPUS", "TOTAL_PHYMEM", "BOOT_TIME",
+    "version_info", "__version__",
+    "STATUS_RUNNING", "STATUS_IDLE", "STATUS_SLEEPING", "STATUS_DISK_SLEEP",
+    "STATUS_STOPPED", "STATUS_TRACING_STOP", "STATUS_ZOMBIE", "STATUS_DEAD",
+    "STATUS_WAKING", "STATUS_LOCKED",
     # classes
-    "Process",
+    "Process", "Popen",
     # functions
     "test", "pid_exists", "get_pid_list", "process_iter", "get_process_list",
     "avail_phymem", "used_phymem", "total_virtmem", "avail_virtmem",
@@ -29,12 +33,18 @@ import time
 import signal
 import warnings
 import errno
+import subprocess
 try:
     import pwd
 except ImportError:
     pwd = None
 
-from psutil.error import Error, NoSuchProcess, AccessDenied
+from psutil.error import Error, NoSuchProcess, AccessDenied, TimeoutExpired
+from psutil._compat import property
+from psutil._common import (STATUS_RUNNING, STATUS_IDLE, STATUS_SLEEPING,
+                            STATUS_DISK_SLEEP, STATUS_STOPPED,
+                            STATUS_TRACING_STOP, STATUS_ZOMBIE, STATUS_DEAD,
+                            STATUS_WAKING, STATUS_LOCKED)
 
 # import the appropriate module for our platform only
 if sys.platform.lower().startswith("linux"):
@@ -43,6 +53,9 @@ if sys.platform.lower().startswith("linux"):
 
 elif sys.platform.lower().startswith("win32"):
     from psutil._psmswindows import *
+    __all__.extend(["ABOVE_NORMAL_PRIORITY_CLASS", "BELOW_NORMAL_PRIORITY_CLASS",
+                    "HIGH_PRIORITY_CLASS", "IDLE_PRIORITY_CLASS",
+                    "NORMAL_PRIORITY_CLASS", "REALTIME_PRIORITY_CLASS"])
 
 elif sys.platform.lower().startswith("darwin"):
     from psutil._psosx import *
@@ -54,44 +67,6 @@ else:
     raise NotImplementedError('platform %s is not supported' % sys.platform)
 
 
-class CPUTimes:
-    """This class contains information about CPU times.
-    It is not used directly but it's returned as an instance by
-    psutil.cpu_times() function.
-
-    Every CPU time is accessible in form of an attribute and represents
-    the time CPU has spent in the given mode.
-
-    The attributes availability varies depending on the platform.
-    Here follows a list of all available attributes:
-
-     - user
-     - system
-     - idle
-     - nice (UNIX)
-     - iowait (Linux)
-     - irq (Linux, FreeBSD)
-     - softirq (Linux)
-    """
-
-    def __init__(self, **kwargs):
-        self.__attrs = []
-        for name in kwargs:
-            setattr(self, name, kwargs[name])
-            self.__attrs.append(name)
-
-    def __str__(self):
-        string = []
-        for attr in self.__attrs:
-            value = getattr(self, attr)
-            string.append("%s=%s" %(attr, value))
-        return '; '.join(string)
-
-    def __iter__(self):
-        for attr in self.__attrs:
-            yield getattr(self, attr)
-
-
 class Process(object):
     """Represents an OS process."""
 
@@ -124,8 +99,8 @@ class Process(object):
                 details = "(pid=%s, name=%s, cmdline=%s)" % (pid, name, cmdline)
             else:
                 details = "(pid=%s, name=%s)" % (pid, name)
-        return "%s.%s %s" % (self.__class__.__module__,
-                             self.__class__.__name__, details)
+        return "%s.%s%s" % (self.__class__.__module__,
+                            self.__class__.__name__, details)
 
     def __repr__(self):
         return "<%s at %s>" % (self.__str__(), id(self))
@@ -135,8 +110,12 @@ class Process(object):
         and creation time.
         """
         h1 = (self.pid, self.create_time)
-        h2 = (other.pid, other.create_time)
-        return h1 == h2
+        try:
+            h2 = (other.pid, other.create_time)
+        except AttributeError:
+            return False
+        else:
+            return h1 == h2
 
     @property
     def pid(self):
@@ -150,11 +129,15 @@ class Process(object):
 
     @property
     def parent(self):
-        """Return the parent process as a Process object. If no ppid is
-        known then return None."""
-        if self.ppid is not None:
-            return Process(self.ppid)
-        return None
+        """Return the parent process as a Process object. If no parent
+        pid is known return None.
+        """
+        ppid = self.ppid
+        if ppid is not None:
+            try:
+                return Process(ppid)
+            except NoSuchProcess:
+                pass
 
     @property
     def name(self):
@@ -185,6 +168,8 @@ class Process(object):
                 _exe = os.path.realpath(cmdline[0])
                 if os.path.isfile(_exe) and os.access(_exe, os.X_OK):
                     return _exe
+        if not exe:
+            raise AccessDenied(self.pid, self._platform_impl._process_name)
         return exe
 
     @property
@@ -198,24 +183,65 @@ class Process(object):
         """The command line process has been called with."""
         return self._platform_impl.get_process_cmdline()
 
+    @property
+    def status(self):
+        """The process current status as a STATUS_* constant."""
+        return self._platform_impl.get_process_status()
+
+    @property
+    def nice(self):
+        """Get or set process niceness (priority)."""
+        return self._platform_impl.get_process_nice()
+
+    @nice.setter
+    def nice(self, value):
+        # invoked on "p.nice = num"; change process niceness
+        return self._platform_impl.set_process_nice(value)
+
+    if os.name == 'posix':
+
+        @property
+        def uids(self):
+            """Return a named tuple denoting the process real,
+            effective, and saved user ids.
+            """
+            return self._platform_impl.get_process_uids()
+
+        @property
+        def gids(self):
+            """Return a named tuple denoting the process real,
+            effective, and saved group ids.
+            """
+            return self._platform_impl.get_process_gids()
+
     @property
     def uid(self):
-        """The real user id of the current process."""
-        return self._platform_impl.get_process_uid()
+        """The real user id of the current process (deprecated)."""
+        warnings.warn("'uid' property is deprecated; use 'uids.real' instead",
+                      DeprecationWarning)
+        if os.name != 'posix':
+            return -1
+        return self.uids.real
 
     @property
     def gid(self):
-        """The real group id of the current process."""
-        return self._platform_impl.get_process_gid()
+        """The real group id of the current process (deprecated)."""
+        warnings.warn("'gid' property is deprecated; use 'uids.real' instead",
+                      DeprecationWarning)
+        if os.name != 'posix':
+            return -1
+        return self.gids.real
 
     @property
     def username(self):
-        """The name of the user that owns the process."""
+        """The name of the user that owns the process.
+        On UNIX this is calculated by using *real* process uid.
+        """
         if os.name == 'posix':
             if pwd is None:
-                # might happen on compiled-from-sources python
+                # might happen if python was installed from sources
                 raise ImportError("requires pwd module shipped with standard python")
-            return pwd.getpwuid(self.uid).pw_name
+            return pwd.getpwuid(self.uids.real).pw_name
         else:
             return self._platform_impl.get_process_username()
 
@@ -228,16 +254,48 @@ class Process(object):
 
     # available for Windows and Linux only
     if hasattr(PlatformProcess, "get_process_cwd"):
+
         def getcwd(self):
             """Return a string representing the process current working
             directory.
             """
             return self._platform_impl.get_process_cwd()
 
+    # Linux, BSD and Windows only
+    if hasattr(PlatformProcess, "get_process_io_counters"):
+
+        def get_io_counters(self):
+            """Return process I/O statistics as a namedtuple including
+            the number of read/write calls performed and the amount of
+            bytes read and written by the process.
+            """
+            return self._platform_impl.get_process_io_counters()
+
+    # available only on Linux
+    if hasattr(PlatformProcess, "get_process_ionice"):
+
+        def get_ionice(self):
+            """Return process I/O niceness (priority) as a namedtuple."""
+            return self._platform_impl.get_process_ionice()
+
+        def set_ionice(self, ioclass, value=None):
+            """Set process I/O niceness (priority).
+            ioclass is one of the IOPRIO_CLASS_* constants.
+            iodata is a number which goes from 0 to 7. The higher the
+            value, the lower the I/O priority of the process.
+            """
+            return self._platform_impl.set_process_ionice(ioclass, value)
+
     def get_num_threads(self):
         """Return the number of threads used by this process."""
         return self._platform_impl.get_process_num_threads()
 
+    def get_threads(self):
+        """Return threads opened by process as a list of namedtuples
+        including thread id and thread CPU times (user/system).
+        """
+        return self._platform_impl.get_process_threads()
+
     def get_children(self):
         """Return the children of this process as a list of Process
         objects.
@@ -304,8 +362,7 @@ class Process(object):
 
     def get_cpu_times(self):
         """Return a tuple whose values are process CPU user and system
-        time.  These are the same first two values that os.times()
-        returns for the current process.
+        times. The same as os.times() but per-process.
         """
         return self._platform_impl.get_cpu_times()
 
@@ -332,15 +389,15 @@ class Process(object):
 
     def get_open_files(self):
         """Return files opened by process as a list of namedtuples
-        including absolute file name and file descriptor.
+        including absolute file name and file descriptor number.
         """
         return self._platform_impl.get_open_files()
 
     def get_connections(self):
         """Return TCP and UPD connections opened by process as a list
-        of namedtuple/s.
-        For third party processes (!= os.getpid()) results can differ
-        depending on user privileges.
+        of namedtuples.
+        On BSD and OSX results for third party processes (!= os.getpid())
+        can differ depending on user privileges.
         """
         return self._platform_impl.get_connections()
 
@@ -426,14 +483,65 @@ class Process(object):
         else:
             self._platform_impl.kill_process()
 
+    def wait(self, timeout=None):
+        """Wait for process to terminate and, if process is a children
+        of the current one also return its exit code, else None.
+        """
+        return self._platform_impl.process_wait(timeout)
+
+
+class Popen(Process):
+    """A more convenient interface to stdlib subprocess module.
+    It starts a sub process and deals with it exactly as when using
+    subprocess.Popen class but in addition also provides all the
+    property and methods of psutil.Process class in a unique interface:
+
+      >>> import psutil
+      >>> from subprocess import PIPE
+      >>> p = psutil.Popen(["/usr/bin/python", "-c", "print 'hi'"], stdout=PIPE)
+      >>> p.name
+      'python'
+      >>> p.uids
+      user(real=1000, effective=1000, saved=1000)
+      >>> p.username
+      'giampaolo'
+      >>> p.communicate()
+      ('hi\n', None)
+      >>> p.terminate()
+      >>> p.wait(timeout=2)
+      0
+      >>>
+
+    For method names common to both classes such as kill(), terminate()
+    and wait(), psutil.Process implementation takes precedence.
+
+    For a complete documentation refers to:
+    http://docs.python.org/library/subprocess.html
+    """
+
+    def __init__(self, *args, **kwargs):
+        self.__subproc = subprocess.Popen(*args, **kwargs)
+        Process.__init__(self, self.__subproc.pid)
+
+    def __dir__(self):
+        return list(set(dir(Popen) + dir(subprocess.Popen)))
+
+    def __getattribute__(self, name):
+        try:
+            return object.__getattribute__(self, name)
+        except AttributeError:
+            try:
+                return object.__getattribute__(self.__subproc, name)
+            except AttributeError:
+                raise AttributeError("%s instance has no attribute '%s'"
+                                      %(self.__class__.__name__, name))
+
 
 def process_iter():
     """Return an iterator yielding a Process class instances for all
     running processes on the local machine.
     """
     pids = get_pid_list()
-    # for each PID, create a proxyied Process object
-    # it will lazy init it's name and path later if required
     for pid in pids:
         try:
             yield Process(pid)
@@ -447,9 +555,19 @@ def get_process_list():
     return list(process_iter())
 
 def cpu_times():
-    """Return system CPU times as a CPUTimes object."""
-    values = get_system_cpu_times()
-    return CPUTimes(**values)
+    """Return system CPU times as a namedtuple object.
+    Every CPU time represents the time CPU has spent in the given mode.
+    The attributes availability varies depending on the platform.
+    Here follows a list of all available attributes:
+     - user
+     - system
+     - idle
+     - nice (UNIX)
+     - iowait (Linux)
+     - irq (Linux, FreeBSD)
+     - softirq (Linux)
+    """
+    return get_system_cpu_times()
 
 
 _last_cpu_times = cpu_times()
