@@ -3,14 +3,6 @@
 # $Id$
 #
 
-__all__ = ["NUM_CPUS", "TOTAL_PHYMEM",
-           "PlatformProcess",
-           "avail_phymem", "used_phymem", "total_virtmem", "avail_virtmem",
-           "used_virtmem", "get_system_cpu_times", "pid_exists", "get_pid_list",
-           "phymem_buffers", "cached_phymem"
-          ]
-
-
 import os
 import errno
 import socket
@@ -18,16 +10,22 @@ import struct
 import sys
 import base64
 
-try:
-    from collections import namedtuple
-except ImportError:
-    from psutil.compat import namedtuple  # python < 2.6
-
+import _psutil_posix
+import _psutil_linux
 from psutil import _psposix
-from psutil.error import AccessDenied, NoSuchProcess
+from psutil.error import AccessDenied, NoSuchProcess, TimeoutExpired
+from psutil._compat import namedtuple
+from psutil._common import *
+
+# Linux specific extended namespace
+__all__ = base_module_namespace[:]
+__all__.extend([
+    "IOPRIO_CLASS_NONE", "IOPRIO_CLASS_RT", "IOPRIO_CLASS_BE", "IOPRIO_CLASS_IDLE",
+    "phymem_buffers", "cached_phymem",
+    ])
 
 
-def _get_uptime():
+def _get_boot_time():
     """Return system boot time (epoch in seconds)"""
     f = open('/proc/stat', 'r')
     for line in f:
@@ -56,11 +54,16 @@ def _get_total_phymem():
 
 # Number of clock ticks per second
 _CLOCK_TICKS = os.sysconf(os.sysconf_names["SC_CLK_TCK"])
-_UPTIME = _get_uptime()
+BOOT_TIME = _get_boot_time()
 NUM_CPUS = _get_num_cpus()
 TOTAL_PHYMEM = _get_total_phymem()
+# ioprio_* constants http://linux.die.net/man/2/ioprio_get
+IOPRIO_CLASS_NONE = 0
+IOPRIO_CLASS_RT = 1
+IOPRIO_CLASS_BE = 2
+IOPRIO_CLASS_IDLE = 3
 
-del _get_uptime, _get_num_cpus, _get_total_phymem
+del _get_boot_time, _get_num_cpus, _get_total_phymem
 
 # http://students.mimuw.edu.pl/lxr/source/include/net/tcp_states.h
 _TCP_STATES_TABLE = {"01" : "ESTABLISHED",
@@ -147,9 +150,7 @@ def get_system_cpu_times():
 
     values = values[1:8]
     values = tuple([float(x) / _CLOCK_TICKS for x in values])
-
-    return dict(user=values[0], nice=values[1], system=values[2], idle=values[3],
-                iowait=values[4], irq=values[5], softirq=values[6])
+    return ntuple_sys_cputimes(*values[:7])
 
 def get_pid_list():
     """Returns a list of PIDs currently running on the system."""
@@ -163,6 +164,18 @@ def pid_exists(pid):
     return _psposix.pid_exists(pid)
 
 
+# taken from /fs/proc/array.c
+_status_map = {"R" : STATUS_RUNNING,
+               "S" : STATUS_SLEEPING,
+               "D" : STATUS_DISK_SLEEP,
+               "T" : STATUS_STOPPED,
+               "t" : STATUS_TRACING_STOP,
+               "Z" : STATUS_ZOMBIE,
+               "X" : STATUS_DEAD,
+               "x" : STATUS_DEAD,
+               "K" : STATUS_WAKE_KILL,
+               "W" : STATUS_WAKING}
+
 # --- decorators
 
 def wrap_exceptions(callable):
@@ -173,7 +186,10 @@ def wrap_exceptions(callable):
         try:
             return callable(self, *args, **kwargs)
         except (OSError, IOError), err:
-            if err.errno == errno.ENOENT:  # no such file or directory
+            # ENOENT (no such file or directory) gets raised on open().
+            # ESRCH (no such process) can get raised on read() if
+            # process is gone in meantime.
+            if err.errno in (errno.ENOENT, errno.ESRCH):
                 raise NoSuchProcess(self.pid, self._process_name)
             if err.errno in (errno.EPERM, errno.EACCES):
                 raise AccessDenied(self.pid, self._process_name)
@@ -184,11 +200,6 @@ def wrap_exceptions(callable):
 class LinuxProcess(object):
     """Linux process implementation."""
 
-    _meminfo_ntuple = namedtuple('meminfo', 'rss vms')
-    _cputimes_ntuple = namedtuple('cputimes', 'user system')
-    _openfile_ntuple = namedtuple('openfile', 'path fd')
-    _connection_ntuple = namedtuple('connection', 'fd family type local_address '
-                                                  'remote_address status')
     __slots__ = ["pid", "_process_name"]
 
     def __init__(self, pid):
@@ -209,7 +220,7 @@ class LinuxProcess(object):
 
     def get_process_exe(self):
         if self.pid in (0, 2):
-            return ""   # special case for kernel processes
+            raise AccessDenied(self.pid, self._process_name)
         try:
             exe = os.readlink("/proc/%s/exe" % self.pid)
         except (OSError, IOError), err:
@@ -226,6 +237,10 @@ class LinuxProcess(object):
                 raise AccessDenied(self.pid, self._process_name)
             raise
 
+        # readlink() might return paths containing null bytes causing
+        # problems when used with other fs-related functions (os.*,
+        # open(), ...)
+        exe = exe.replace('\x00', '')
         # It seems symlinks can point to a deleted/invalid location
         # (this usually  happens with "pulseaudio" process).
         # However, if we had permissions to execute readlink() it's
@@ -246,11 +261,28 @@ class LinuxProcess(object):
         finally:
             f.close()
 
+    @wrap_exceptions
+    def get_process_io_counters(self):
+        # special case for 0 (kernel process) PID
+        if self.pid == 0:
+            return ntuple_io(0, 0, 0, 0)
+        f = open("/proc/%s/io" % self.pid)
+        for line in f:
+            if line.startswith("rchar"):
+                read_count = int(line.split()[1])
+            elif line.startswith("wchar"):
+                write_count = int(line.split()[1])
+            elif line.startswith("read_bytes"):
+                read_bytes = int(line.split()[1])
+            elif line.startswith("write_bytes"):
+                write_bytes = int(line.split()[1])
+        return ntuple_io(read_count, write_count, read_bytes, write_bytes)
+
     @wrap_exceptions
     def get_cpu_times(self):
         # special case for 0 (kernel process) PID
         if self.pid == 0:
-            return self._cputimes_ntuple(0.0, 0.0)
+            return ntuple_cputimes(0.0, 0.0)
         f = open("/proc/%s/stat" % self.pid)
         st = f.read().strip()
         f.close()
@@ -259,13 +291,20 @@ class LinuxProcess(object):
         values = st.split(' ')
         utime = float(values[11]) / _CLOCK_TICKS
         stime = float(values[12]) / _CLOCK_TICKS
-        return self._cputimes_ntuple(utime, stime)
+        return ntuple_cputimes(utime, stime)
+
+    @wrap_exceptions
+    def process_wait(self, timeout=None):
+        try:
+            return _psposix.wait_pid(self.pid, timeout)
+        except TimeoutExpired:
+            raise TimeoutExpired(self.pid, self._process_name)
 
     @wrap_exceptions
     def get_process_create_time(self):
         # special case for 0 (kernel processes) PID; return system uptime
         if self.pid == 0:
-            return _UPTIME
+            return BOOT_TIME
         f = open("/proc/%s/stat" % self.pid)
         st = f.read().strip()
         f.close()
@@ -276,14 +315,14 @@ class LinuxProcess(object):
         # unit is jiffies (clock ticks).
         # We first divide it for clock ticks and then add uptime returning
         # seconds since the epoch, in UTC.
-        starttime = (float(values[19]) / _CLOCK_TICKS) + _UPTIME
+        starttime = (float(values[19]) / _CLOCK_TICKS) + BOOT_TIME
         return starttime
 
     @wrap_exceptions
     def get_memory_info(self):
         # special case for 0 (kernel processes) PID
         if self.pid == 0:
-            return self._meminfo_ntuple(0, 0)
+            return ntuple_meminfo(0, 0)
         f = open("/proc/%s/status" % self.pid)
         virtual_size = 0
         resident_size = 0
@@ -296,13 +335,17 @@ class LinuxProcess(object):
                 resident_size = int(line.split()[1]) * 1024
                 break
         f.close()
-        return self._meminfo_ntuple(resident_size, virtual_size)
+        return ntuple_meminfo(resident_size, virtual_size)
 
     @wrap_exceptions
     def get_process_cwd(self):
         if self.pid == 0:
-            return ''
-        return os.readlink("/proc/%s/cwd" % self.pid)
+            raise AccessDenied(self.pid, self._process_name)
+        # readlink() might return paths containing null bytes causing
+        # problems when used with other fs-related functions (os.*,
+        # open(), ...)
+        path = os.readlink("/proc/%s/cwd" % self.pid)
+        return path.replace('\x00', '')
 
     @wrap_exceptions
     def get_process_num_threads(self):
@@ -314,6 +357,88 @@ class LinuxProcess(object):
                 f.close()
                 return int(line.split()[1])
 
+    @wrap_exceptions
+    def get_process_threads(self):
+        if self.pid == 0:
+            return []
+        thread_ids = os.listdir("/proc/%s/task" % self.pid)
+        thread_ids.sort()
+        retlist = []
+        for thread_id in thread_ids:
+            try:
+                f = open("/proc/%s/task/%s/stat" % (self.pid, thread_id))
+            except (OSError, IOError), err:
+                if err.errno == errno.ENOENT:
+                    # no such file or directory; it means thread
+                    # disappeared on us
+                    continue
+                raise
+            st = f.read().strip()
+            f.close()
+            # ignore the first two values ("pid (exe)")
+            st = st[st.find(')') + 2:]
+            values = st.split(' ')
+            utime = float(values[11]) / _CLOCK_TICKS
+            stime = float(values[12]) / _CLOCK_TICKS
+            ntuple = ntuple_thread(int(thread_id), utime, stime)
+            retlist.append(ntuple)
+        return retlist
+
+    @wrap_exceptions
+    def get_process_nice(self):
+        #f = open('/proc/%s/stat' % self.pid, 'r')
+        #data = f.read()
+        #f.close()
+        #return int(data.split()[18])
+
+        # Use C implementation
+        return _psutil_posix.getpriority(self.pid)
+
+    @wrap_exceptions
+    def set_process_nice(self, value):
+        return _psutil_posix.setpriority(self.pid, value)
+
+    # only starting from kernel 2.6.13
+    if hasattr(_psutil_linux, "ioprio_get"):
+
+        @wrap_exceptions
+        def get_process_ionice(self):
+            ioclass, value = _psutil_linux.ioprio_get(self.pid)
+            return ntuple_ionice(ioclass, value)
+
+        @wrap_exceptions
+        def set_process_ionice(self, ioclass, value):
+            if ioclass in (IOPRIO_CLASS_NONE, None):
+                if value:
+                    raise ValueError("can't specify value with IOPRIO_CLASS_NONE")
+                ioclass = IOPRIO_CLASS_NONE
+                value = 0
+            if ioclass in (IOPRIO_CLASS_RT, IOPRIO_CLASS_BE):
+                if value is None:
+                    value = 4
+            elif ioclass == IOPRIO_CLASS_IDLE:
+                if value:
+                    raise ValueError("can't specify value with IOPRIO_CLASS_IDLE")
+                value = 0
+            else:
+                value = 0
+            if not 0 <= value <= 8:
+                raise ValueError("value argument range expected is between 0 and 8")
+            return _psutil_linux.ioprio_set(self.pid, ioclass, value)
+
+    @wrap_exceptions
+    def get_process_status(self):
+        if self.pid == 0:
+            return 0
+        f = open("/proc/%s/status" % self.pid)
+        for line in f:
+            if line.startswith("State:"):
+                f.close()
+                letter = line.split()[1]
+                if letter in _status_map:
+                    return _status_map[letter]
+                return constant(-1, '?')
+
     @wrap_exceptions
     def get_open_files(self):
         retlist = []
@@ -329,16 +454,10 @@ class LinuxProcess(object):
                 if file == "[]":
                     continue
                 if os.path.isfile(file) and not file in retlist:
-                    ntuple = self._openfile_ntuple(file, int(fd))
+                    ntuple = ntuple_openfile(file, int(fd))
                     retlist.append(ntuple)
         return retlist
 
-#    --- lsof implementation
-#
-#    def get_open_files(self):
-#        lsof = _psposix.LsofParser(self.pid, self._process_name)
-#        return lsof.get_process_open_files()
-
     @wrap_exceptions
     def get_connections(self):
         if self.pid == 0:
@@ -375,8 +494,8 @@ class LinuxProcess(object):
                     else:
                         status = ""
                     fd = int(inodes[inode])
-                    conn = self._connection_ntuple(fd, family, _type, laddr,
-                                                   raddr, status)
+                    conn = ntuple_connection(fd, family, _type, laddr,
+                                             raddr, status)
                     retlist.append(conn)
             f.close()
             return retlist
@@ -405,30 +524,26 @@ class LinuxProcess(object):
                 return int(line.split()[1])
 
     @wrap_exceptions
-    def get_process_uid(self):
+    def get_process_uids(self):
         if self.pid == 0:
-            return 0
+            return ntuple_uids(0, 0, 0)
         f = open("/proc/%s/status" % self.pid)
         for line in f:
             if line.startswith('Uid:'):
-                # Uid line provides 4 values which stand for real,
-                # effective, saved set, and file system UIDs.
-                # We want to provide real UID only.
                 f.close()
-                return int(line.split()[1])
+                _, real, effective, saved, fs = line.split()
+                return ntuple_uids(int(real), int(effective), int(saved))
 
     @wrap_exceptions
-    def get_process_gid(self):
+    def get_process_gids(self):
         if self.pid == 0:
-            return 0
+            return ntuple_uids(0, 0, 0)
         f = open("/proc/%s/status" % self.pid)
         for line in f:
             if line.startswith('Gid:'):
-                # Uid line provides 4 values which stand for real,
-                # effective, saved set, and file system GIDs.
-                # We want to provide real GID only.
                 f.close()
-                return int(line.split()[1])
+                _, real, effective, saved, fs = line.split()
+                return ntuple_gids(int(real), int(effective), int(saved))
 
     @staticmethod
     def _decode_address(addr, family):
