@@ -6,19 +6,20 @@
 import errno
 import os
 
-try:
-    from collections import namedtuple
-except ImportError:
-    from psutil.compat import namedtuple  # python < 2.6
-
 import _psutil_osx
+import _psutil_posix
 import _psposix
-from psutil.error import AccessDenied, NoSuchProcess
+from psutil.error import AccessDenied, NoSuchProcess, TimeoutExpired
+from psutil._compat import namedtuple
+from psutil._common import *
+
+__all__ = base_module_namespace[:]
 
 # --- constants
 
 NUM_CPUS = _psutil_osx.get_num_cpus()
 TOTAL_PHYMEM = _psutil_osx.get_total_phymem()
+BOOT_TIME = _psutil_osx.get_system_boot_time()
 
 # --- functions
 
@@ -42,11 +43,11 @@ def used_virtmem():
     """Return the amount of used memory currently in use on the system, in bytes."""
     return _psutil_osx.get_total_virtmem() - _psutil_osx.get_avail_virtmem()
 
+_cputimes_ntuple = namedtuple('cputimes', 'user nice system idle')
 def get_system_cpu_times():
-    """Return a dict representing the following CPU times:
-    user, nice, system, idle."""
-    values = _psutil_osx.get_system_cpu_times()
-    return dict(user=values[0], nice=values[1], system=values[2], idle=values[3])
+    """Return system CPU times as a namedtuple."""
+    user, nice, system, idle = _psutil_osx.get_system_cpu_times()
+    return _cputimes_ntuple(user, nice, system, idle)
 
 def get_pid_list():
     """Returns a list of PIDs currently running on the system."""
@@ -75,11 +76,17 @@ def wrap_exceptions(callable):
     return wrapper
 
 
+_status_map = {
+    _psutil_osx.SIDL : STATUS_IDLE,
+    _psutil_osx.SRUN : STATUS_RUNNING,
+    _psutil_osx.SSLEEP : STATUS_SLEEPING,
+    _psutil_osx.SSTOP : STATUS_STOPPED,
+    _psutil_osx.SZOMB : STATUS_ZOMBIE,
+}
+
 class OSXProcess(object):
     """Wrapper class around underlying C implementation."""
 
-    _meminfo_ntuple = namedtuple('meminfo', 'rss vms')
-    _cputimes_ntuple = namedtuple('cputimes', 'user system')
     __slots__ = ["pid", "_process_name"]
 
     def __init__(self, pid):
@@ -111,25 +118,26 @@ class OSXProcess(object):
         return _psutil_osx.get_process_ppid(self.pid)
 
     @wrap_exceptions
-    def get_process_uid(self):
-        """Return process real user id."""
-        return _psutil_osx.get_process_uid(self.pid)
+    def get_process_uids(self):
+        real, effective, saved = _psutil_osx.get_process_uids(self.pid)
+        return ntuple_uids(real, effective, saved)
+
 
     @wrap_exceptions
-    def get_process_gid(self):
-        """Return process real group id."""
-        return _psutil_osx.get_process_gid(self.pid)
+    def get_process_gids(self):
+        real, effective, saved = _psutil_osx.get_process_gids(self.pid)
+        return ntuple_gids(real, effective, saved)
 
     @wrap_exceptions
     def get_memory_info(self):
         """Return a tuple with the process' RSS and VMS size."""
         rss, vms = _psutil_osx.get_memory_info(self.pid)
-        return self._meminfo_ntuple(rss, vms)
+        return ntuple_meminfo(rss, vms)
 
     @wrap_exceptions
     def get_cpu_times(self):
         user, system = _psutil_osx.get_cpu_times(self.pid)
-        return self._cputimes_ntuple(user, system)
+        return ntuple_cputimes(user, system)
 
     @wrap_exceptions
     def get_process_create_time(self):
@@ -142,16 +150,59 @@ class OSXProcess(object):
         """Return the number of threads belonging to the process."""
         return _psutil_osx.get_process_num_threads(self.pid)
 
+    @wrap_exceptions
     def get_open_files(self):
-        """Return files opened by process by parsing lsof output."""
-        lsof = _psposix.LsofParser(self.pid, self._process_name)
-        return lsof.get_process_open_files()
+        """Return files opened by process."""
+        if self.pid == 0:
+            raise AccessDenied(self.pid, self._process_name)
+        files = []
+        rawlist = _psutil_osx.get_process_open_files(self.pid)
+        for path, fd in rawlist:
+            if os.path.isfile(path):
+                ntuple = ntuple_openfile(path, fd)
+                files.append(ntuple)
+        return files
 
+    @wrap_exceptions
     def get_connections(self):
         """Return etwork connections opened by a process as a list of
         namedtuples."""
-        lsof = _psposix.LsofParser(self.pid, self._process_name)
-        return lsof.get_process_connections()
+        retlist = _psutil_osx.get_process_connections(self.pid)
+        return [ntuple_connection(*conn) for conn in retlist]
+
+    @wrap_exceptions
+    def process_wait(self, timeout=None):
+        try:
+            return _psposix.wait_pid(self.pid, timeout)
+        except TimeoutExpired:
+            raise TimeoutExpired(self.pid, self._process_name)
+
+    @wrap_exceptions
+    def get_process_nice(self):
+        return _psutil_posix.getpriority(self.pid)
+
+    @wrap_exceptions
+    def set_process_nice(self, value):
+        return _psutil_posix.setpriority(self.pid, value)
+
+    @wrap_exceptions
+    def get_process_status(self):
+        code = _psutil_osx.get_process_status(self.pid)
+        if code in _status_map:
+            return _status_map[code]
+        return constant(-1, "?")
+
+    @wrap_exceptions
+    def get_process_threads(self):
+        """Return the number of threads belonging to the process."""
+        rawlist = _psutil_osx.get_process_threads(self.pid)
+        retlist = []
+        for thread_id, utime, stime in rawlist:
+            ntuple = ntuple_thread(thread_id, utime, stime)
+            retlist.append(ntuple)
+        return retlist
+
+
 
 PlatformProcess = OSXProcess
 
