@@ -23,12 +23,14 @@ import warnings
 import atexit
 import errno
 import threading
+import tempfile
 
 import psutil
 
 
 PYTHON = os.path.realpath(sys.executable)
 DEVNULL = open(os.devnull, 'r+')
+TESTFN = os.path.join(os.getcwd(), "$testfile")
 
 POSIX = os.name == 'posix'
 LINUX = sys.platform.lower().startswith("linux")
@@ -59,6 +61,18 @@ def get_test_subprocess(cmd=None, stdout=DEVNULL, stderr=DEVNULL, stdin=None):
     _subprocesses_started.add(sproc.pid)
     return sproc
 
+def sh(cmdline):
+    """run cmd in a subprocess and return its output.
+    raises RuntimeError on error.
+    """
+    p = subprocess.Popen(cmdline, shell=True, stdout=subprocess.PIPE,
+                                              stderr=subprocess.PIPE)
+    stdout, stderr = p.communicate()
+    if p.returncode != 0:
+        raise RuntimeError(stderr)
+    if stderr:
+        warnings.warn(stderr, RuntimeWarning)
+
 def wait_for_pid(pid, timeout=1):
     """Wait for pid to show up in the process list then return.
     Used in the test suite to give time the sub process to initialize.
@@ -73,38 +87,11 @@ def wait_for_pid(pid, timeout=1):
         if time.time() >= raise_at:
             raise RuntimeError("Timed out")
 
-def kill(pid):
-    """Kill a process given its PID."""
-    if hasattr(os, 'kill'):
-        os.kill(pid, signal.SIGKILL)
-    else:
-        psutil.Process(pid).kill()
-
 def reap_children(search_all=False):
     """Kill any subprocess started by this test suite and ensure that
     no zombies stick around to hog resources and create problems when
     looking for refleaks.
     """
-    if POSIX:
-        def waitpid(process):
-            # on posix we are free to wait for any pending process by
-            # passing -1 to os.waitpid()
-            while True:
-                try:
-                    any_process = -1
-                    pid, status = os.waitpid(any_process, os.WNOHANG)
-                    if pid == 0 and not process.is_running():
-                        break
-                except OSError:
-                    if not process.is_running():
-                        break
-    else:
-        def waitpid(process):
-            # on non-posix systems we just wait for the given process
-            # to go away
-            while process.is_running():
-                time.sleep(0.01)
-
     if search_all:
         this_process = psutil.Process(os.getpid())
         pids = [x.pid for x in this_process.get_children()]
@@ -118,7 +105,8 @@ def reap_children(search_all=False):
         except psutil.NoSuchProcess:
             pass
         else:
-            waitpid(child)
+            child.wait()
+
 
 # we want to search through all processes before exiting
 atexit.register(reap_children, search_all=True)
@@ -155,6 +143,54 @@ def skipUnless(condition, reason="", warn=False):
         return skipIf(True, reason, warn)
     return skipIf(False)
 
+def ignore_access_denied(fun):
+    """Decorator to Ignore AccessDenied exceptions."""
+    def outer(fun, *args, **kwargs):
+        def inner(self):
+            try:
+                return fun(self, *args, **kwargs)
+            except psutil.AccessDenied:
+                pass
+        return inner
+    return outer
+
+
+class ThreadTask(threading.Thread):
+    """A thread object used for running process thread tests."""
+
+    def __init__(self):
+        threading.Thread.__init__(self)
+        self._running = False
+        self._interval = None
+        self._flag = threading.Event()
+
+    def __repr__(self):
+        name =  self.__class__.__name__
+        return '<%s running=%s at %#x>' % (name, self._running, id(self))
+
+    def start(self, interval=0.001):
+        """Start thread and keep it running until an explicit
+        stop() request. Polls for shutdown every 'timeout' seconds.
+        """
+        if self._running:
+            raise ValueError("already started")
+        self._interval = interval
+        threading.Thread.start(self)
+        self._flag.wait()
+
+    def run(self):
+        self._running = True
+        self._flag.set()
+        while self._running:
+            time.sleep(self._interval)
+
+    def stop(self):
+        """Stop thread execution and and waits until it is stopped."""
+        if not self._running:
+            raise ValueError("already stopped")
+        self._running = False
+        self.join()
+
 
 class TestCase(unittest.TestCase):
 
@@ -178,7 +214,7 @@ class TestCase(unittest.TestCase):
         p = psutil.Process(test_pid)
         name = p.name
         p.kill()
-        sproc.wait()
+        p.wait()
         self.assertFalse(psutil.pid_exists(test_pid) and name == PYTHON)
 
     def test_terminate(self):
@@ -188,7 +224,7 @@ class TestCase(unittest.TestCase):
         p = psutil.Process(test_pid)
         name = p.name
         p.terminate()
-        sproc.wait()
+        p.wait()
         self.assertFalse(psutil.pid_exists(test_pid) and name == PYTHON)
 
     def test_send_signal(self):
@@ -201,14 +237,85 @@ class TestCase(unittest.TestCase):
         p = psutil.Process(test_pid)
         name = p.name
         p.send_signal(sig)
-        sproc.wait()
+        p.wait()
         self.assertFalse(psutil.pid_exists(test_pid) and name == PYTHON)
 
+    def test_wait(self):
+        # check exit code signal
+        sproc = get_test_subprocess()
+        p = psutil.Process(sproc.pid)
+        p.kill()
+        code = p.wait()
+        if os.name == 'posix':
+            self.assertEqual(code, signal.SIGKILL)
+        else:
+            self.assertEqual(code, 0)
+        self.assertFalse(p.is_running())
+
+        sproc = get_test_subprocess()
+        p = psutil.Process(sproc.pid)
+        p.terminate()
+        code = p.wait()
+        if os.name == 'posix':
+            self.assertEqual(code, signal.SIGTERM)
+        else:
+            self.assertEqual(code, 0)
+        self.assertFalse(p.is_running())
+
+        # check sys.exit() code
+        code = "import time, sys; time.sleep(0.01); sys.exit(5);"
+        sproc = get_test_subprocess([PYTHON, "-c", code])
+        p = psutil.Process(sproc.pid)
+        self.assertEqual(p.wait(), 5)
+        self.assertFalse(p.is_running())
+
+        # Test wait() issued twice.
+        # It is not supposed to raise NSP when the process is gone.
+        # On UNIX this should return None, on Windows it should keep
+        # returning the exit code.
+        sproc = get_test_subprocess([PYTHON, "-c", code])
+        p = psutil.Process(sproc.pid)
+        self.assertEqual(p.wait(), 5)
+        self.assertTrue(p.wait() in (5, None))
+
+        # test timeout
+        sproc = get_test_subprocess()
+        p = psutil.Process(sproc.pid)
+        p.name
+        self.assertRaises(psutil.TimeoutExpired, p.wait, 0.01)
+
+    @skipUnless(POSIX)
+    def test_wait_non_children(self):
+        # test wait() against processes which are not our children
+        code = "import sys;"
+        code += "from subprocess import Popen, PIPE;"
+        code += "cmd = ['%s', '-c', 'import time; time.sleep(10)'];" %PYTHON
+        code += "sp = Popen(cmd, stdout=PIPE);"
+        code += "sys.stdout.write(str(sp.pid));"
+        sproc = get_test_subprocess([PYTHON, "-c", code], stdout=subprocess.PIPE)
+
+        grandson_pid = int(sproc.stdout.read())
+        grandson_proc = psutil.Process(grandson_pid)
+        try:
+            self.assertRaises(psutil.TimeoutExpired, grandson_proc.wait, 0.01)
+            grandson_proc.kill()
+            ret = grandson_proc.wait()
+            self.assertEqual(ret, None)
+        finally:
+            if grandson_proc.is_running():
+                grandson_proc.kill()
+                grandson_proc.wait()
+
     def test_TOTAL_PHYMEM(self):
         x = psutil.TOTAL_PHYMEM
         self.assertTrue(isinstance(x, (int, long)))
         self.assertTrue(x > 0)
 
+    def test_BOOT_TIME(self):
+        x = psutil.BOOT_TIME
+        self.assertTrue(isinstance(x, float))
+        self.assertTrue(x > 0)
+
     def test_used_phymem(self):
         x = psutil.used_phymem()
         self.assertTrue(isinstance(x, (int, long)))
@@ -249,13 +356,11 @@ class TestCase(unittest.TestCase):
     def test_system_cpu_times(self):
         total = 0
         times = psutil.cpu_times()
-        self.assertTrue(isinstance(times, psutil.CPUTimes))
         sum(times)
         for cp_time in times:
             self.assertTrue(isinstance(cp_time, float))
             self.assertTrue(cp_time >= 0.0)
             total += cp_time
-        # test CPUTimes's __iter__ and __str__ implementation
         self.assertEqual(total, sum(times))
         str(times)
 
@@ -331,20 +436,107 @@ class TestCase(unittest.TestCase):
         # make sure returned value can be pretty printed with strftime
         time.strftime("%Y %m %d %H:%M:%S", time.localtime(p.create_time))
 
-    def test_get_num_threads(self):
+    @skipIf(OSX, warn=False)
+    def test_get_io_counters(self):
         p = psutil.Process(os.getpid())
-        numt1 = p.get_num_threads()
-        if not WINDOWS and not OSX:
-            # test is unreliable on Windows and OS X
-            # NOTE: sleep(1) is too long for OS X, works with sleep(.5)
-            self.assertEqual(numt1, 1)
-        t = threading.Thread(target=lambda:time.sleep(1))
-        t.start()
-        numt2 = p.get_num_threads()
-        if WINDOWS:
-           self.assertTrue(numt2 > numt1)
+        # test reads
+        io1 = p.get_io_counters()
+        f = open(PYTHON, 'rb')
+        f.read()
+        f.close()
+        io2 = p.get_io_counters()
+        if not BSD:
+            self.assertTrue(io2.read_count > io1.read_count)
+            self.assertTrue(io2.write_count == io1.write_count)
+        self.assertTrue(io2.read_bytes >= io1.read_bytes)
+        self.assertTrue(io2.write_bytes >= io1.write_bytes)
+        # test writes
+        io1 = p.get_io_counters()
+        f = tempfile.TemporaryFile()
+        if sys.version_info >= (3,):
+            f.write(bytes("x" * 1000000, 'ascii'))
         else:
-            self.assertEqual(numt2, 2)
+            f.write("x" * 1000000)
+        f.close()
+        io2 = p.get_io_counters()
+        if not BSD:
+            self.assertTrue(io2.write_count > io1.write_count)
+            self.assertTrue(io2.write_bytes > io1.write_bytes)
+        self.assertTrue(io2.read_count >= io1.read_count)
+        self.assertTrue(io2.read_bytes >= io1.read_bytes)
+
+    @skipUnless(LINUX)
+    def test_get_set_ionice(self):
+        from psutil import (IOPRIO_CLASS_NONE, IOPRIO_CLASS_RT, IOPRIO_CLASS_BE,
+                            IOPRIO_CLASS_IDLE)
+        self.assertEqual(IOPRIO_CLASS_NONE, 0)
+        self.assertEqual(IOPRIO_CLASS_RT, 1)
+        self.assertEqual(IOPRIO_CLASS_BE, 2)
+        self.assertEqual(IOPRIO_CLASS_IDLE, 3)
+        p = psutil.Process(os.getpid())
+        try:
+            p.set_ionice(2)
+            ioclass, value = p.get_ionice()
+            self.assertEqual(ioclass, 2)
+            self.assertEqual(value, 4)
+            #
+            p.set_ionice(3)
+            ioclass, value = p.get_ionice()
+            self.assertEqual(ioclass, 3)
+            self.assertEqual(value, 0)
+            #
+            p.set_ionice(2, 0)
+            ioclass, value = p.get_ionice()
+            self.assertEqual(ioclass, 2)
+            self.assertEqual(value, 0)
+            p.set_ionice(2, 7)
+            ioclass, value = p.get_ionice()
+            self.assertEqual(ioclass, 2)
+            self.assertEqual(value, 7)
+            self.assertRaises(ValueError, p.set_ionice, 2, 10)
+        finally:
+            p.set_ionice(IOPRIO_CLASS_NONE)
+
+    def test_get_num_threads(self):
+        # on certain platforms such as Linux we might test for exact
+        # thread number, since we always have with 1 thread per process,
+        # but this does not apply across all platforms (OSX, Windows)
+        p = psutil.Process(os.getpid())
+        step1 = p.get_num_threads()
+
+        thread = ThreadTask()
+        thread.start()
+        try:
+            step2 = p.get_num_threads()
+            self.assertEqual(step2, step1 + 1)
+            thread.stop()
+        finally:
+            if thread._running:
+                thread.stop()
+
+    def test_get_threads(self):
+        p = psutil.Process(os.getpid())
+        step1 = p.get_threads()
+
+        thread = ThreadTask()
+        thread.start()
+
+        try:
+            step2 = p.get_threads()
+            self.assertEqual(len(step2), len(step1) + 1)
+            # on Linux, first thread id is supposed to be this process
+            if LINUX:
+                self.assertEqual(step2[0].id, os.getpid())
+            athread = step2[0]
+            # test named tuple
+            self.assertEqual(athread.id, athread[0])
+            self.assertEqual(athread.user_time, athread[1])
+            self.assertEqual(athread.system_time, athread[2])
+            # test num threads
+            thread.stop()
+        finally:
+            if thread._running:
+                thread.stop()
 
     def test_get_memory_info(self):
         p = psutil.Process(os.getpid())
@@ -384,23 +576,35 @@ class TestCase(unittest.TestCase):
         wait_for_pid(sproc.pid)
         p = psutil.Process(sproc.pid)
         self.assertTrue(p.is_running())
-        psutil.Process(sproc.pid).kill()
-        sproc.wait()
+        p.kill()
+        p.wait()
         self.assertFalse(p.is_running())
 
     def test_pid_exists(self):
         sproc = get_test_subprocess()
         wait_for_pid(sproc.pid)
         self.assertTrue(psutil.pid_exists(sproc.pid))
-        psutil.Process(sproc.pid).kill()
-        sproc.wait()
+        p = psutil.Process(sproc.pid)
+        p.kill()
+        p.wait()
         self.assertFalse(psutil.pid_exists(sproc.pid))
         self.assertFalse(psutil.pid_exists(-1))
 
     def test_exe(self):
         sproc = get_test_subprocess()
         wait_for_pid(sproc.pid)
-        self.assertEqual(psutil.Process(sproc.pid).exe, PYTHON)
+        try:
+            self.assertEqual(psutil.Process(sproc.pid).exe, PYTHON)
+        except AssertionError:
+            # certain platforms such as BSD are more accurate returning:
+            # "/usr/local/bin/python2.7"
+            # ...instead of:
+            # "/usr/local/bin/python"
+            # We do not want to consider this difference in accuracy
+            # an error.
+            extended_py_name = PYTHON + "%s.%s" % (sys.version_info.major,
+                                                   sys.version_info.minor)
+            self.assertEqual(psutil.Process(sproc.pid).exe, extended_py_name)
         for p in psutil.process_iter():
             try:
                 exe = p.exe
@@ -437,35 +641,76 @@ class TestCase(unittest.TestCase):
         else:
             self.assertEqual(psutil.Process(sproc.pid).name, os.path.basename(PYTHON))
 
-    def test_uid(self):
-        sproc = get_test_subprocess()
-        wait_for_pid(sproc.pid)
-        uid = psutil.Process(sproc.pid).uid
-        if hasattr(os, 'getuid'):
-            self.assertEqual(uid, os.getuid())
-        else:
-            # On those platforms where UID doesn't make sense (Windows)
-            # we expect it to be -1
-            self.assertEqual(uid, -1)
+    if os.name == 'posix':
 
-    def test_gid(self):
-        sproc = get_test_subprocess()
-        wait_for_pid(sproc.pid)
-        gid = psutil.Process(sproc.pid).gid
-        if hasattr(os, 'getgid'):
-            self.assertEqual(gid, os.getgid())
-        else:
-            # On those platforms where GID doesn't make sense (Windows)
-            # we expect it to be -1
-            self.assertEqual(gid, -1)
+        def test_uids(self):
+            p = psutil.Process(os.getpid())
+            real, effective, saved = p.uids
+            # os.getuid() refers to "real" uid
+            self.assertEqual(real, os.getuid())
+            # os.geteuid() refers to "effective" uid
+            self.assertEqual(effective, os.geteuid())
+            # no such thing as os.getsuid() ("saved" uid), but starting
+            # from python 2.7 we have os.getresuid()[2]
+            if hasattr(os, "getresuid"):
+                self.assertEqual(saved, os.getresuid()[2])
+
+        def test_gids(self):
+            p = psutil.Process(os.getpid())
+            real, effective, saved = p.gids
+            # os.getuid() refers to "real" uid
+            self.assertEqual(real, os.getgid())
+            # os.geteuid() refers to "effective" uid
+            self.assertEqual(effective, os.getegid())
+            # no such thing as os.getsuid() ("saved" uid), but starting
+            # from python 2.7 we have os.getresgid()[2]
+            if hasattr(os, "getresuid"):
+                self.assertEqual(saved, os.getresgid()[2])
+
+        def test_nice(self):
+            p = psutil.Process(os.getpid())
+            self.assertRaises(TypeError, setattr, p, "nice", "str")
+            first_nice = p.nice
+            try:
+                p.nice = 1
+                self.assertEqual(p.nice, 1)
+                # going back to previous nice value raises AccessDenied on OSX
+                if not OSX:
+                    p.nice = 0
+                    self.assertEqual(p.nice, 0)
+            finally:
+                # going back to previous nice value raises AccessDenied on OSX
+                if not OSX:
+                    p.nice = first_nice
+
+    if os.name == 'nt':
+
+        def test_nice(self):
+            p = psutil.Process(os.getpid())
+            self.assertRaises(TypeError, setattr, p, "nice", "str")
+            try:
+                self.assertEqual(p.nice, psutil.NORMAL_PRIORITY_CLASS)
+                p.nice = psutil.HIGH_PRIORITY_CLASS
+                self.assertEqual(p.nice, psutil.HIGH_PRIORITY_CLASS)
+                p.nice = psutil.NORMAL_PRIORITY_CLASS
+                self.assertEqual(p.nice, psutil.NORMAL_PRIORITY_CLASS)
+            finally:
+                p.nice = psutil.NORMAL_PRIORITY_CLASS
+
+    def test_status(self):
+        p = psutil.Process(os.getpid())
+        self.assertEqual(p.status, psutil.STATUS_RUNNING)
+        self.assertEqual(str(p.status), "running")
+        for p in psutil.process_iter():
+            if str(p.status) == '?':
+                self.fail("invalid status for pid %d" % p.pid)
 
     def test_username(self):
         sproc = get_test_subprocess()
         p = psutil.Process(sproc.pid)
         if POSIX:
             import pwd
-            user = pwd.getpwuid(p.uid).pw_name
-            self.assertEqual(p.username, user)
+            self.assertEqual(p.username, pwd.getpwuid(os.getuid()).pw_name)
         elif WINDOWS:
             expected_username = os.environ['USERNAME']
             expected_domain = os.environ['USERDOMAIN']
@@ -493,27 +738,26 @@ class TestCase(unittest.TestCase):
         self.assertEqual(p.getcwd(), expected_dir)
 
     def test_get_open_files(self):
-        thisfile = os.path.join(os.getcwd(), __file__)
-
         # current process
         p = psutil.Process(os.getpid())
         files = p.get_open_files()
-        self.assertFalse(thisfile in files)
-        f = open(thisfile, 'r')
+        self.assertFalse(TESTFN in files)
+        f = open(TESTFN, 'r')
+        time.sleep(.1)
         filenames = [x.path for x in p.get_open_files()]
-        self.assertTrue(thisfile in filenames)
+        self.assertTrue(TESTFN in filenames)
         f.close()
         for file in filenames:
             self.assertTrue(os.path.isfile(file))
 
         # another process
-        cmdline = "import time; f = open(r'%s', 'r'); time.sleep(100);" % thisfile
+        cmdline = "import time; f = open(r'%s', 'r'); time.sleep(100);" % TESTFN
         sproc = get_test_subprocess([PYTHON, "-c", cmdline])
         wait_for_pid(sproc.pid)
         time.sleep(0.1)
         p = psutil.Process(sproc.pid)
         filenames = [x.path for x in p.get_open_files()]
-        self.assertTrue(thisfile in filenames)
+        self.assertTrue(TESTFN in filenames)
         for file in filenames:
             self.assertTrue(os.path.isfile(file))
         # all processes
@@ -528,7 +772,7 @@ class TestCase(unittest.TestCase):
 
     def test_get_open_files2(self):
         # test fd and path fields
-        fileobj = open(os.path.join(os.getcwd(), __file__), 'r')
+        fileobj = open(TESTFN, 'r')
         p = psutil.Process(os.getpid())
         for path, fd in p.get_open_files():
             if path == fileobj.name or fd == fileobj.fileno():
@@ -665,9 +909,9 @@ class TestCase(unittest.TestCase):
             tcp6_proc = None
             udp6_proc = None
 
+        # --- check connections of all processes
+
         time.sleep(0.1)
-        this_proc = psutil.Process(os.getpid())
-        children_pids = [p.pid for p in this_proc.get_children()]
         for p in psutil.process_iter():
             try:
                 cons = p.get_connections()
@@ -701,41 +945,43 @@ class TestCase(unittest.TestCase):
                                 continue
                             else:
                                 raise
-                        self.assertEqual(dupsock.family, conn.family)
-                        self.assertEqual(dupsock.type, conn.type)
+                        # python >= 2.5
+                        if hasattr(dupsock, "family"):
+                            self.assertEqual(dupsock.family, conn.family)
+                            self.assertEqual(dupsock.type, conn.type)
 
-                # check matches against subprocesses
-                if p.pid in children_pids:
-                    self.assertTrue(len(cons) == 1)
-                    conn = cons[0]
-                    # TCP v4
-                    if p.pid == tcp4_proc.pid:
-                        self.assertEqual(conn.family, socket.AF_INET)
-                        self.assertEqual(conn.type, socket.SOCK_STREAM)
-                        self.assertEqual(conn.local_address[0], "127.0.0.1")
-                        self.assertEqual(conn.remote_address, ())
-                        self.assertEqual(conn.status, "LISTEN")
-                    # UDP v4
-                    elif p.pid == udp4_proc.pid:
-                        self.assertEqual(conn.family, socket.AF_INET)
-                        self.assertEqual(conn.type, socket.SOCK_DGRAM)
-                        self.assertEqual(conn.local_address[0], "127.0.0.1")
-                        self.assertEqual(conn.remote_address, ())
-                        self.assertEqual(conn.status, "")
-                    # TCP v6
-                    elif p.pid == getattr(tcp6_proc, "pid", None):
-                        self.assertEqual(conn.family, socket.AF_INET6)
-                        self.assertEqual(conn.type, socket.SOCK_STREAM)
-                        self.assertEqual(conn.local_address[0], "::1")
-                        self.assertEqual(conn.remote_address, ())
-                        self.assertEqual(conn.status, "LISTEN")
-                    # UDP v6
-                    elif p.pid == getattr(udp6_proc, "pid", None):
-                        self.assertEqual(conn.family, socket.AF_INET6)
-                        self.assertEqual(conn.type, socket.SOCK_DGRAM)
-                        self.assertEqual(conn.local_address[0], "::1")
-                        self.assertEqual(conn.remote_address, ())
-                        self.assertEqual(conn.status, "")
+        # --- check matches against subprocesses
+
+        for p in psutil.Process(os.getpid()).get_children():
+            for conn in p.get_connections():
+                # TCP v4
+                if p.pid == tcp4_proc.pid:
+                    self.assertEqual(conn.family, socket.AF_INET)
+                    self.assertEqual(conn.type, socket.SOCK_STREAM)
+                    self.assertEqual(conn.local_address[0], "127.0.0.1")
+                    self.assertEqual(conn.remote_address, ())
+                    self.assertEqual(conn.status, "LISTEN")
+                # UDP v4
+                elif p.pid == udp4_proc.pid:
+                    self.assertEqual(conn.family, socket.AF_INET)
+                    self.assertEqual(conn.type, socket.SOCK_DGRAM)
+                    self.assertEqual(conn.local_address[0], "127.0.0.1")
+                    self.assertEqual(conn.remote_address, ())
+                    self.assertEqual(conn.status, "")
+                # TCP v6
+                elif p.pid == getattr(tcp6_proc, "pid", None):
+                    self.assertEqual(conn.family, socket.AF_INET6)
+                    self.assertEqual(conn.type, socket.SOCK_STREAM)
+                    self.assertEqual(conn.local_address[0], "::1")
+                    self.assertEqual(conn.remote_address, ())
+                    self.assertEqual(conn.status, "LISTEN")
+                # UDP v6
+                elif p.pid == getattr(udp6_proc, "pid", None):
+                    self.assertEqual(conn.family, socket.AF_INET6)
+                    self.assertEqual(conn.type, socket.SOCK_DGRAM)
+                    self.assertEqual(conn.local_address[0], "::1")
+                    self.assertEqual(conn.remote_address, ())
+                    self.assertEqual(conn.status, "")
 
     def test_parent_ppid(self):
         this_parent = os.getpid()
@@ -762,7 +1008,11 @@ class TestCase(unittest.TestCase):
         sproc = get_test_subprocess()
         p = psutil.Process(sproc.pid)
         p.suspend()
+        time.sleep(0.1)
+        self.assertEqual(p.status, psutil.STATUS_STOPPED)
+        self.assertEqual(str(p.status), "stopped")
         p.resume()
+        self.assertTrue(p.status != psutil.STATUS_STOPPED)
 
     def test_get_pid_list(self):
         plist = [x.pid for x in psutil.get_process_list()]
@@ -780,51 +1030,6 @@ class TestCase(unittest.TestCase):
         finally:
             sys.stdout = stdout
 
-    def test_types(self):
-        sproc = get_test_subprocess()
-        wait_for_pid(sproc.pid)
-        p = psutil.Process(sproc.pid)
-        self.assert_(isinstance(p.pid, int))
-        self.assert_(isinstance(p.ppid, int))
-        self.assert_(isinstance(p.parent, psutil.Process))
-        self.assert_(isinstance(p.name, str))
-        if self.__class__.__name__ != "LimitedUserTestCase":
-            self.assert_(isinstance(p.exe, str))
-        self.assert_(isinstance(p.cmdline, list))
-        self.assert_(isinstance(p.uid, int))
-        self.assert_(isinstance(p.gid, int))
-        self.assert_(isinstance(p.create_time, float))
-        self.assert_(isinstance(p.username, (unicode, str)))
-        if hasattr(p, 'getcwd'):
-            if not POSIX and self.__class__.__name__ != "LimitedUserTestCase":
-                self.assert_(isinstance(p.getcwd(), str))
-        if not POSIX and self.__class__.__name__ != "LimitedUserTestCase":
-            self.assert_(isinstance(p.get_open_files(), list))
-            for path, fd in p.get_open_files():
-                self.assert_(isinstance(path, (unicode, str)))
-                self.assert_(isinstance(fd, int))
-        if not POSIX and self.__class__.__name__ != "LimitedUserTestCase" \
-        and SUPPORT_CONNECTIONS:
-            self.assert_(isinstance(p.get_connections(), list))
-        self.assert_(isinstance(p.is_running(), bool))
-        if not OSX or self.__class__.__name__ != "LimitedUserTestCase":
-            self.assert_(isinstance(p.get_cpu_times(), tuple))
-            self.assert_(isinstance(p.get_cpu_times()[0], float))
-            self.assert_(isinstance(p.get_cpu_times()[1], float))
-            self.assert_(isinstance(p.get_cpu_percent(0), float))
-            self.assert_(isinstance(p.get_memory_info(), tuple))
-            self.assert_(isinstance(p.get_memory_info()[0], int))
-            self.assert_(isinstance(p.get_memory_info()[1], int))
-            self.assert_(isinstance(p.get_memory_percent(), float))
-            self.assert_(isinstance(p.get_num_threads(), int))
-        self.assert_(isinstance(psutil.get_process_list(), list))
-        self.assert_(isinstance(psutil.get_process_list()[0], psutil.Process))
-        self.assert_(isinstance(psutil.process_iter(), types.GeneratorType))
-        self.assert_(isinstance(psutil.process_iter().next(), psutil.Process))
-        self.assert_(isinstance(psutil.get_pid_list(), list))
-        self.assert_(isinstance(psutil.get_pid_list()[0], int))
-        self.assert_(isinstance(psutil.pid_exists(1), bool))
-
     def test_invalid_pid(self):
         self.assertRaises(ValueError, psutil.Process, "1")
         self.assertRaises(ValueError, psutil.Process, None)
@@ -832,7 +1037,7 @@ class TestCase(unittest.TestCase):
         self.assertRaises(psutil.NoSuchProcess, psutil.Process, -1)
 
     def test_zombie_process(self):
-        # Test that NoSuchProcess exception gets raised in the event the
+        # Test that NoSuchProcess exception gets raised in case the
         # process dies after we create the Process object.
         # Example:
         #  >>> proc = Process(1234)
@@ -842,19 +1047,36 @@ class TestCase(unittest.TestCase):
         sproc = get_test_subprocess()
         p = psutil.Process(sproc.pid)
         p.kill()
-        sproc.wait()
+        p.wait()
 
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "ppid")
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "parent")
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "name")
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "exe")
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "cmdline")
-        self.assertRaises(psutil.NoSuchProcess, getattr, p, "uid")
-        self.assertRaises(psutil.NoSuchProcess, getattr, p, "gid")
+        self.assertRaises(psutil.NoSuchProcess, getattr, p, "status")
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "create_time")
         self.assertRaises(psutil.NoSuchProcess, getattr, p, "username")
         if hasattr(p, 'getcwd'):
             self.assertRaises(psutil.NoSuchProcess, p.getcwd)
+        if POSIX:
+            self.assertRaises(psutil.NoSuchProcess, getattr, p, "uids")
+            self.assertRaises(psutil.NoSuchProcess, getattr, p, "gids")
+        self.assertRaises(psutil.NoSuchProcess, getattr, p, "nice")
+        try:
+            if os.name == 'posix':
+                p.nice = 1
+            else:
+                p.nice = psutil.NORMAL_PRIORITY_CLASS
+        except psutil.NoSuchProcess:
+            pass
+        else:
+            self.fail("exception not raised")
+        if hasattr(p, 'get_ionice'):
+            self.assertRaises(psutil.NoSuchProcess, p.get_ionice)
+            self.assertRaises(psutil.NoSuchProcess, p.set_ionice, 2)
+        if hasattr(p, 'get_io_counters'):
+            self.assertRaises(psutil.NoSuchProcess, p.get_io_counters)
         self.assertRaises(psutil.NoSuchProcess, p.get_open_files)
         self.assertRaises(psutil.NoSuchProcess, p.get_connections)
         self.assertRaises(psutil.NoSuchProcess, p.suspend)
@@ -868,6 +1090,7 @@ class TestCase(unittest.TestCase):
         self.assertRaises(psutil.NoSuchProcess, p.get_memory_percent)
         self.assertRaises(psutil.NoSuchProcess, p.get_children)
         self.assertRaises(psutil.NoSuchProcess, p.get_num_threads)
+        self.assertRaises(psutil.NoSuchProcess, p.get_threads)
         self.assertFalse(p.is_running())
 
     def test__str__(self):
@@ -878,36 +1101,58 @@ class TestCase(unittest.TestCase):
         sproc = get_test_subprocess()
         p = psutil.Process(sproc.pid)
         p.kill()
-        sproc.wait()
+        p.wait()
         self.assertTrue(str(sproc.pid) in str(p))
         self.assertTrue("terminated" in str(p))
 
     def test_fetch_all(self):
         valid_procs = 0
-        attrs = ['__str__', 'create_time', 'username', 'getcwd', 'get_cpu_times',
-                 'get_memory_info', 'get_memory_percent', 'get_open_files',
-                  'get_num_threads']
+        excluded_names = ['send_signal', 'suspend', 'resume', 'terminate',
+                          'kill', 'wait']
+        excluded_names += ['get_cpu_percent', 'path', 'uid', 'gid', 'get_children']
+        # XXX - skip slow lsof implementation;
+        if BSD:
+           excluded_names += ['get_open_files', 'get_connections']
+        if OSX:
+           excluded_names += ['get_connections']
+        attrs = []
+        for name in dir(psutil.Process):
+            if name.startswith("_"):
+                continue
+            if name.startswith("set_"):
+                continue
+            if name in excluded_names:
+                continue
+            attrs.append(name)
+
         for p in psutil.process_iter():
-            for attr in attrs:
-                # skip slow Python implementation; we're reasonably sure
-                # it works anyway
-                if POSIX and attr == 'get_open_files':
-                    continue
+            for name in attrs:
                 try:
-                    attr = getattr(p, attr, None)
-                    if attr is not None and callable(attr):
-                        attr()
-                    valid_procs += 1
-                except (psutil.NoSuchProcess, psutil.AccessDenied), err:
-                    self.assertEqual(err.pid, p.pid)
-                    if err.name:
-                        self.assertEqual(err.name, p.name)
-                    self.assertTrue(str(err))
-                    self.assertTrue(err.msg)
-                except:
+                    try:
+                        attr = getattr(p, name, None)
+                        if attr is not None and callable(attr):
+                            ret = attr()
+                        else:
+                            ret = attr
+                        valid_procs += 1
+                    except (psutil.NoSuchProcess, psutil.AccessDenied), err:
+                        self.assertEqual(err.pid, p.pid)
+                        if err.name:
+                            self.assertEqual(err.name, p.name)
+                        self.assertTrue(str(err))
+                        self.assertTrue(err.msg)
+                    else:
+                        if name == 'parent' or ret in (0, 0.0, []):
+                            continue
+                        self.assertTrue(ret)
+                        if name == "exe":
+                            self.assertTrue(os.path.isfile(ret))
+                        elif name == "getcwd":
+                            self.assertTrue(os.path.isdir(ret))
+                except Exception, err:
                     trace = traceback.format_exc()
-                    self.fail('Exception raised for method %s, pid %s:\n%s'
-                              %(attr, p.pid, trace))
+                    self.fail('%s\nmethod=%s, pid=%s, retvalue=%s'
+                              %(trace, name, p.pid, repr(ret)))
 
         # we should always have a non-empty list, not including PID 0 etc.
         # special cases.
@@ -917,28 +1162,17 @@ class TestCase(unittest.TestCase):
         # Process(0) is supposed to work on all platforms even if with
         # some differences
         p = psutil.Process(0)
-        if WINDOWS:
-            self.assertEqual(p.name, 'System Idle Process')
-        elif LINUX:
-            self.assertEqual(p.name, 'sched')
-        elif BSD:
-            self.assertEqual(p.name, 'swapper')
-        elif OSX:
-            self.assertEqual(p.name, 'kernel_task')
+        self.assertTrue(p.name)
 
         if os.name == 'posix':
-            self.assertEqual(p.uid, 0)
-            self.assertEqual(p.gid, 0)
-        else:
-            self.assertEqual(p.uid, -1)
-            self.assertEqual(p.gid, -1)
+            self.assertEqual(p.uids.real, 0)
+            self.assertEqual(p.gids.real, 0)
 
         self.assertTrue(p.ppid in (0, 1))
-        self.assertEqual(p.exe, "")
+        #self.assertEqual(p.exe, "")
         self.assertEqual(p.cmdline, [])
-        # this can either raise AD (Win) or return 0 (UNIX)
         try:
-            self.assertTrue(p.get_num_threads() in (0, 1))
+            p.get_num_threads()
         except psutil.AccessDenied:
             pass
 
@@ -960,7 +1194,19 @@ class TestCase(unittest.TestCase):
         self.assertTrue(0 in psutil.get_pid_list())
         self.assertTrue(psutil.pid_exists(0))
 
-    # --- OS specific tests
+    def test_Popen(self):
+        # Popen class test
+        cmd = [PYTHON, "-c", "import time; time.sleep(3600);"]
+        proc = psutil.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        try:
+            proc.name
+            proc.stdin
+            self.assertTrue(hasattr(proc, 'name'))
+            self.assertTrue(hasattr(proc, 'stdin'))
+            self.assertRaises(AttributeError, getattr, proc, 'foo')
+        finally:
+            proc.kill()
+            proc.wait()
 
 
 if hasattr(os, 'getuid'):
@@ -987,6 +1233,14 @@ if hasattr(os, 'getuid'):
             # DeprecationWarning is only raised once
             pass
 
+        def test_nice(self):
+            try:
+                psutil.Process(os.getpid()).nice = -1
+            except psutil.AccessDenied:
+                pass
+            else:
+                self.fail("exception not raised")
+
         # overridden tests known to raise AccessDenied when run
         # as limited user on different platforms
 
@@ -1018,14 +1272,20 @@ if hasattr(os, 'getuid'):
             def test_get_connections(self):
                 self.assertRaises(psutil.AccessDenied, TestCase.test_get_connections, self)
 
+            def test_get_connections_all(self):
+                self.assertRaises(psutil.AccessDenied, TestCase.test_get_connections_all, self)
+
             def test_connection_fromfd(self):
                 self.assertRaises(psutil.AccessDenied, TestCase.test_connection_fromfd, self)
 
+        if OSX:
+            def test_nice(self):
+                self.assertRaises(psutil.AccessDenied, TestCase.test_nice, self)
+
 
 def test_main():
     tests = []
     test_suite = unittest.TestSuite()
-
     tests.append(TestCase)
 
     if hasattr(os, 'getuid'):
@@ -1053,6 +1313,10 @@ def test_main():
     for test_class in tests:
         test_suite.addTest(unittest.makeSuite(test_class))
 
+    f = open(TESTFN, 'w')
+    f.close()
+    atexit.register(lambda: os.remove(TESTFN))
+
     unittest.TextTestRunner(verbosity=2).run(test_suite)
     DEVNULL.close()
 
